﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Design;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Wisej.Base;
using Wisej.Core;
using Wisej.Core.Design;

namespace Wisej.Web.Ext.Uploader
{
	/// <summary>
	/// CountUp.js is a dependency-free, lightweight JavaScript widget that can be used to 
	/// create animations that display numerical data in a more interesting way.
	/// https://github.com/inorganik/countUp.js.
	/// </summary>
	[ToolboxItem(true)]
	[ToolboxBitmap(typeof(Uploader))]
	[Description("CountUp.js is a JavaScript widget that can create animations to display numerical data.")]
	public class Uploader : Button, IWisejControl
	{
        private CancellationTokenSource CancelToken;

		/// <summary>
		/// Renders the client component.
		/// </summary>
		/// <param name="config">Dynamic configuration object.</param>
		protected override void OnWebRender(dynamic config)
		{
			base.OnWebRender((object)config);

			config.className = "wisej.web.ext.Uploader";
            config.webMethods = new[] { "UploadFileError", "UploadFileSuccess", "UploadFileCancelled", "UploadFileChanged", "UpdateFileState" };           
        }

        private ConcurrentQueue<FileChunk> fileChunks = new ConcurrentQueue<FileChunk>();
        private object myLock = new object();
        public event CancelEventHandler FileOk;
        public event EventHandler OnCancelled;
        public event EventHandler OnStarted;
        public event EventHandler<UploaderProgressEventArgs> UploadProgress;
        public event EventHandler<UploaderErrorEventArgs> Error;
        public event EventHandler OnFileReady;
        private SpeedMonitor UploadMonitor = new SpeedMonitor();
        private bool IsRunning;
        public Uploader()
        {
            if (!string.IsNullOrEmpty(Application.StartupPath))
            {
                this.TempPath = Path.Combine(Application.StartupPath, "Temp");

                if (!Directory.Exists(TempPath)) Directory.CreateDirectory(TempPath);
            }
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public UploadFileInfo File
        {
            get
            {
                return Files.FirstOrDefault();
            }
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public List<UploadFileInfo> Files { get; set; } = new List<UploadFileInfo>();
        private UploadFileInfo UploadFile { get; set; }
        #region Wisej Implementation
        public string TempPath { get; }
        public string FileName { get; private set; }
        public string FilePath { get; private set; }
        public long FileSize { get; private set; }
        public long BytesDone { get; private set; }
        public string PartialFilePath { get; private set; }      

        #endregion

        [WebMethod]
        public void UploadFileError(string error)
        {
            Error?.Invoke(this, new UploaderErrorEventArgs(UploadErrorType.Upload, error, new string[] { FileName }, new long[] { FileSize }));
        }

        [WebMethod]
        public void UploadFileSuccess()
        {
            var file = new UploadFileInfo()
            {
                OriginalFileName = FileName,
                FilePath = PartialFilePath,
                Type = this.UploadFile.Type,
                FileSize = FileSize
            };
            Files.Add(file);

            UpdateProgress();
            IsRunning = false;

            Application.Update(this, () =>
            {
                CancelEventArgs args = new CancelEventArgs();
                this.FileOk?.Invoke(this, args);
                if (args.Cancel)
                {
                    file.Delete();
                }
            });
        }
        
        [WebMethod]
        public void UpdateFileState(double percentComplete)
        {
            lock (myLock)
            {
                var current = BytesDone;
                double bytesDone = (double)FileSize * (double)(percentComplete * 100);
                if (bytesDone < current)
                {
                    return;
                }
                BytesDone = (long)bytesDone;
                var diff = BytesDone - current;
                UploadMonitor.AddDelta(diff);
            }
        }

        public void RunInTask(Action action)
        {
            Application.RunInContext(this, () =>
            {
                Application.StartTask(() =>
                {
                    action.Invoke();
                    Application.Update(this);
                });
            });
        }

        [WebMethod]
        public void UploadFileCancelled()
        {
            IsRunning = false;
            RunInTask(() => OnCancelled?.Invoke(this, EventArgs.Empty));
        }

        [WebMethod]
        public void UploadFileChanged(string partialFileName, string fileName, long fileSize, string type)
        {
            if (fileName == null)
            {
                this.UploadFile = null;
                RunInTask(() => OnFileReady?.Invoke(this, EventArgs.Empty));
                return;
            }
            this.UploadFile = new UploadFileInfo()
            {
                OriginalFileName = fileName,
                FileSize = fileSize,
                Type = type,
                TempPath = TempPath
            };
            this.PartialFilePath = Path.Combine(this.TempPath, partialFileName);
            this.FileName = fileName;
            this.FileSize = fileSize;
            RunInTask(() => OnFileReady?.Invoke(this, EventArgs.Empty));
        }

        public class UploadFileInfo
        {
            public string FilePath { get; set; }
            public string OriginalFileName { get; set; }
            public string TempPath { get; set; }
            public long FileSize { get; set; }
            public string Type { get; set; }

            public void Delete()
            {
                if (System.IO.File.Exists(FilePath))
                {
                    System.IO.File.Delete(FilePath);
                }
            }
        }

        public async Task MonitorUpload()
        {
            while (!this.IsDisposed && IsRunning)
            {
                if (this.UploadProgress != null)
                {
                    Application.Update(this, () =>
                    {
                        UpdateProgress();
                    });
                }

                await Task.Delay(500);
            }
        }

        public async Task MonitorUploadSpeed()
        {
            while (!this.IsDisposed && IsRunning)
            {
                UploadMonitor.Tick();

                await Task.Delay(1000);
            }
        }

        private void UpdateProgress()
        {
            this.UploadProgress?.Invoke(this, new UploaderProgressEventArgs(BytesDone, this.FileSize, UploadMonitor.Rate));
        }

        public void Cancel()
        {
            this.CancelToken.Cancel();
            this.Call("cancelDownload");
        }

        public void StartUpload()
        {
            this.Call("startUpload");

            BytesDone = 0;

            CancelToken = new CancellationTokenSource();

            if (!Directory.Exists(TempPath))
            {
                Directory.CreateDirectory(TempPath);
            }

            IsRunning = true;
            _ = MonitorUpload();
            _ = MonitorUploadSpeed();

            RunInTask(() => OnStarted?.Invoke(this, EventArgs.Empty));
        }
    }

    public class UploaderErrorEventArgs : EventArgs
    {
        //
        // Summary:
        //     Initializes a new instance of the Wisej.Web.UploadErrorEventArgs class with the
        //     specified error message.
        //
        // Parameters:
        //   type:
        //     The error type, one of Wisej.Web.UploadErrorType.
        //
        //   message:
        //     The error message.
        //
        //   fileNames:
        //     The list of files that failed to upload.
        //
        //   fileSizes:
        //     The list of files that failed to upload.
        public UploaderErrorEventArgs(UploadErrorType type, string message, string[] fileNames, long[] fileSizes)
        {
            this.ErrorType = type;
            this.Message = message;
            this.FileNames = fileNames;
            this.FileSizes = fileSizes;
        }

        //
        // Summary:
        //     Returns the error type.
        public UploadErrorType ErrorType { get; }
        //
        // Summary:
        //     Returns the error message.
        //
        // Returns:
        //     The error text.
        public string Message { get; }
        //
        // Summary:
        //     List of file names that failed to upload.
        //
        // Returns:
        //     A string array containing the names of the files that failed to upload.
        public string[] FileNames { get; }
        //
        // Summary:
        //     List of file sizes that failed to upload.
        //
        // Returns:
        //     An integer array containing the sizes of the files that failed to upload.
        public long[] FileSizes { get; }
    }

    public class UploaderProgressEventArgs : EventArgs
    {
        //
        // Summary:
        //     Initializes a new instance of the Wisej.Web.UploadProgressEventArgs class with
        //     the specified error message.
        //
        // Parameters:
        //   loaded:
        //     The number of bytes transferred. This doesn't include headers and other overhead,
        //     but only the content itself.
        //
        //   total:
        //     The total number of bytes of content that will be transferred. If the total size
        //     is unknown, this value is zero.
        public UploaderProgressEventArgs(long loaded, long total, int transferRate)
        {
            this.Loaded = loaded;
            this.Total = total;
            this.TransferRate = transferRate;
        }

        //
        // Summary:
        //     The number of bytes transferred.
        public long Loaded { get; }
        //
        // Summary:
        //     The total number of bytes of content that will be transferred.
        public long Total { get; }
        public int TransferRate { get; }
    }

    public class SpeedMonitor
    {
        private const int DefaultAveragePeriod = 12;

        private long total;
        private int speed;
        private int[] speeds;
        private int speedsIndex;
        private DateTime lastUpdated;
        private long tempRecvCount;


        public int Rate
        {
            get { return this.speed; }
        }

        public long Total
        {
            get { return this.total; }
        }


        public SpeedMonitor()
            : this(DefaultAveragePeriod)
        {

        }

        public SpeedMonitor(int averagingPeriod)
        {
            if (averagingPeriod < 0)
                throw new ArgumentOutOfRangeException("averagingPeriod");

            this.lastUpdated = DateTime.UtcNow;
            this.speeds = new int[Math.Max(1, averagingPeriod)];
            this.speedsIndex = -speeds.Length;
        }


        public void AddDelta(int speed)
        {
            this.total += speed;
            this.tempRecvCount += speed;
        }

        public void AddDelta(long speed)
        {
            this.total += speed;
            this.tempRecvCount += speed;
        }

        public void Reset()
        {
            this.total = 0;
            this.speed = 0;
            this.tempRecvCount = 0;
            this.lastUpdated = DateTime.UtcNow;
            this.speedsIndex = -speeds.Length;
        }

        private void TimePeriodPassed(int difference)
        {
            int currSpeed = (int)(tempRecvCount * 1000 / difference);
            tempRecvCount = 0;

            int speedsCount;
            if (speedsIndex < 0)
            {
                //speeds array hasn't been filled yet

                int idx = speeds.Length + speedsIndex;

                speeds[idx] = currSpeed;
                speedsCount = idx + 1;

                speedsIndex++;
            }
            else
            {
                //speeds array is full, keep wrapping around overwriting the oldest value
                speeds[speedsIndex] = currSpeed;
                speedsCount = speeds.Length;

                speedsIndex = (speedsIndex + 1) % speeds.Length;
            }

            int total = speeds[0];
            for (int i = 1; i < speedsCount; i++)
                total += speeds[i];

            speed = total / speedsCount;
        }


        public void Tick()
        {
            DateTime old = lastUpdated;
            DateTime now = DateTime.UtcNow;
            int difference = (int)(now - old).TotalMilliseconds;

            if (difference > 800)
            {
                lastUpdated = now;
                TimePeriodPassed(difference);
            }
        }

        // Used purely for unit testing purposes.
        internal void Tick(int difference)
        {
            lastUpdated = DateTime.UtcNow;
            TimePeriodPassed(difference);
        }
    }
}
